fix some of the bugs related to barrier example...
[model-checker.git] / main.cc
diff --git a/main.cc b/main.cc
index c9ac1d0663206feccb6d05d43803689e7214574e..fc1b2779daa657ae6460293da49d29a5c95ec21d 100644 (file)
--- a/main.cc
+++ b/main.cc
@@ -2,9 +2,11 @@
  *  @brief Entry point for the model checker.
  */
 
-#include "libthreads.h"
+#include <unistd.h>
+
+#include <threads.h>
 #include "common.h"
-#include "threads.h"
+#include "threads-model.h"
 
 #include "datarace.h"
 
 #include "model.h"
 #include "snapshot-interface.h"
 
-/** The thread_system_next function takes the next step in the
- *  execution.  @return Returns the 1 if there is another step and 0
- *  otherwise.
- */
-
-static int thread_system_next(void) {
-       Thread *curr, *next;
-
-       curr = thread_current();
-       if (curr) {
-               if (curr->get_state() == THREAD_READY) {
-                       model->check_current_action();
-                       model->scheduler->add_thread(curr);
-               } else if (curr->get_state() == THREAD_RUNNING)
-                       /* Stopped while running; i.e., completed */
-                       curr->complete();
-               else
-                       ASSERT(false);
-       }
-       next = model->scheduler->next_thread();
-       if (next)
-               next->set_state(THREAD_RUNNING);
-       DEBUG("(%d, %d)\n", curr ? curr->get_id() : -1, next ? next->get_id() : -1);
-       if (!next)
-               return 1;
-       return Thread::swap(model->get_system_context(), next);
+static void param_defaults(struct model_params * params) {
+       params->maxreads = 0;
+       params->maxfuturedelay = 100;
+       params->fairwindow = 0;
+       params->enabledcount = 1;
+       params->bound = 0;
 }
 
-/** The thread_wait_finish method runs the current execution until we
- *  have no more steps to take.
- */
+static void print_usage(struct model_params *params) {
+       printf(
+"Usage: <program name> [MC_OPTIONS] -- [PROGRAM ARGUMENTS]\n"
+"\n"
+"Options:\n"
+"-h                    Display this help message and exit\n"
+"-m                    Maximum times a thread can read from the same write\n"
+"                      while other writes exist. Default: %d\n"
+"-s                    Maximum actions that the model checker will wait for\n"
+"                      a write from the future past the expected number of\n"
+"                      actions. Default: %d\n"
+"-f                    Specify a fairness window in which actions that are\n"
+"                      enabled sufficiently many times should receive\n"
+"                      priority for execution. Default: %d\n"
+"-e                    Enabled count. Default: %d\n"
+"-b                    Upper length bound. Default: %d\n"
+"--                    Program arguments follow.\n\n",
+params->maxreads, params->maxfuturedelay, params->fairwindow, params->enabledcount, params->bound);
+       exit(EXIT_SUCCESS);
+}
 
-static void thread_wait_finish(void) {
-       DBG();
+static void parse_options(struct model_params *params, int *argc, char ***argv) {
+       const char *shortopts = "hm:s:f:e:b:";
+       int opt;
+       bool error = false;
+       while (!error && (opt = getopt(*argc, *argv, shortopts)) != -1) {
+               switch (opt) {
+               case 'h':
+                       print_usage(params);
+                       break;
+               case 's':
+                       params->maxfuturedelay = atoi(optarg);
+                       break;
+               case 'f':
+                       params->fairwindow = atoi(optarg);
+                       break;
+               case 'e':
+                       params->enabledcount = atoi(optarg);
+                       break;
+               case 'b':
+                       params->bound = atoi(optarg);
+                       break;
+               case 'm':
+                       params->maxreads = atoi(optarg);
+                       break;
+               default: /* '?' */
+                       error = true;
+                       break;
+               }
+       }
+       (*argc) -= optind;
+       (*argv) += optind;
 
-       while (!thread_system_next());
+       if (error)
+               print_usage(params);
 }
 
+int main_argc;
+char **main_argv;
 
-/** The real_main function contains the main model checking loop. */
+/** Wrapper to run the user's main function, with appropriate arguments */
+void wrapper_user_main(void *)
+{
+       user_main(main_argc, main_argv);
+}
 
-static void real_main() {
+/** The model_main function contains the main model checking loop. */
+static void model_main() {
        thrd_t user_thread;
-       ucontext_t main_context;
+       struct model_params params;
+
+       param_defaults(&params);
+
+       parse_options(&params, &main_argc, &main_argv);
 
        //Initialize race detector
        initRaceDetector();
@@ -63,21 +103,15 @@ static void real_main() {
        //Create the singleton SnapshotStack object
        snapshotObject = new SnapshotStack();
 
-       model = new ModelChecker();
-
-       if (getcontext(&main_context))
-               return;
-
-       model->set_system_context(&main_context);
+       model = new ModelChecker(params);
 
        snapshotObject->snapshotStep(0);
-
        do {
                /* Start user program */
-               model->add_thread(new Thread(&user_thread, (void (*)(void *)) &user_main, NULL));
+               model->add_thread(new Thread(&user_thread, &wrapper_user_main, NULL));
 
                /* Wait for all threads to complete */
-               thread_wait_finish();
+               model->finish_execution();
        } while (model->next_execution());
 
        delete model;
@@ -85,18 +119,14 @@ static void real_main() {
        DEBUG("Exiting\n");
 }
 
-int main_numargs;
-char ** main_args;
-
 /**
  * Main function.  Just initializes snapshotting library and the
* snapshotting library calls the real_main function.
  */
-int main(int numargs, char ** args) {
-       /* Stash this stuff in case someone wants it eventually */
-       main_numargs=numargs;
-       main_args=args;
+int main(int argc, char ** argv) {
+       main_argc = argc;
+       main_argv = argv;
 
        /* Let's jump in quickly and start running stuff */
-       initSnapShotLibrary(10000, 1024, 1024, 1000, &real_main);
+       initSnapshotLibrary(10000, 1024, 1024, 4000, &model_main);
 }
@@ -85,18 +119,14 @@ static void real_main() {
 /**
  * Main function.  Just initializes snapshotting library and the
- * snapshotting library calls the real_main function.
+ * snapshotting library calls the model_main function.
  */